import ctypes
from .surface import Surface
from sdl2 import *


class Cursor:
    def __init__(self, cursor: SDL_Cursor) -> None:
        self.destroyed = True
        self.cursor = cursor
        self.destroyed = False

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        SDL_FreeCursor(self.cursor)
        self.destroyed = False
        return False

    def __del__(self) -> None:
        self.destroy()


class CursorManager:
    def __init__(self, app: any) -> None:
        self.destroyed = True
        self.app = app
        self.default = Cursor(SDL_GetDefaultCursor())
        self.system = {
            'arrow': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_ARROW)),
            'ibeam': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_IBEAM)),
            'wait': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_WAIT)),
            'crosshair': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_CROSSHAIR)),
            'wait_arrow': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_WAITARROW)),
            'size_nw_se': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZENWSE)),
            'size_ne_sw': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZENESW)),
            'size_we': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZEWE)),
            'size_ns': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZENS)),
            'size_all': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZEALL)),
            'no': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_NO)),
            'hand': Cursor(SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_HAND))
        }
        self.cursor_cache = None
        self.destroyed = False
        # TODO: fix create function

    def set(self, cursor: Cursor = None) -> None:
        self.cursor_cache = cursor
        SDL_SetCursor(cursor and cursor.cursor)

    def create(self, data: any, hot_pos: any = (0, 0)) -> Cursor:
        data_join = self.app.stb(''.join(data))
        result = SDL_CreateCursor(
            (ctypes.c_ubyte * (len(data) * len(data[0])))(
                *data_join.replace(b'X', b'\x01').replace(b'I', b'\x01').replace(b'O', b'\x00').replace(b' ', b'\x00')
            ),
            (ctypes.c_ubyte * (len(data) * len(data[0])))(
                *data_join.replace(b'X', b'\x01').replace(b'I', b'\x00').replace(b'O', b'\x01').replace(b' ', b'\x00')
            ),
            len(data[0]),
            len(data),
            int(hot_pos[0]),
            int(hot_pos[1])
        )
        if not result:
            self.app.raise_error()
        return Cursor(result)

    def from_surface(self, surf: Surface, hot_pos: any = (0, 0)) -> Cursor:
        result = SDL_CreateColorCursor(surf.surface, int(hot_pos[0]), int(hot_pos[1]))
        if not result:
            self.app.raise_error()
        return Cursor(result)

    def get_current(self) -> Cursor:
        result = SDL_GetCursor()
        if not result:
            self.app.raise_error()
        return Cursor(result)

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        del self.cursor_cache
        del self.app
        self.destroyed = False
        return False

    def __del__(self) -> None:
        self.destroy()



from .exceptions import FlagNotFoundError, SDLError
from .app import App
from .clock import Clock, Timer, Animation
from .events import CommonEvent, QuitEvent, AudioDeviceEvent, DropEvent, TouchFingerEvent, KeyboardEvent,\
    MouseMotionEvent, MouseButtonEvent, MouseWheelEvent, TextEditingEvent, TextInputEvent, DisplayEvent, WindowEvent,\
    JoyAxisEvent, JoyBallEvent, JoyButtonEvent, JoyDeviceEvent, JoyHatEvent, JoyBatteryEvent, ControllerAxisEvent,\
    ControllerButtonEvent, ControllerDeviceEvent, ControllerTouchpadEvent, ControllerSensorEvent
from .audio import AudioDeviceManager, AudioSpec, AudioDevice
from .video import BackendManager, Backend, DisplaysManager, Display, DisplayMode, PixelFormat
from .window import Window
from .surface import Surface, SurfaceAnimation
from .texture import Texture
from .renderer import Renderer
from .cursor import CursorManager, Cursor
from .touch import TouchDevice, Finger
from .sensor import Sensor
from .gamecontroller import GameController
from .joystick import Joystick
from .mixer import Mixer, Music, Chunk
from .ttf import TTF
from .loader import Loader
from .math import Math


__version__ = '0.0.3'



import ctypes
from .video import PixelFormat
from sdl2 import *

try:
    from sdl2.sdlimage import *
except:  # noqa
    pass

try:
    IMG_Animation = IMG_Animation
except NameError:
    IMG_Animation = any


class Surface:
    def __init__(self, surf: SDL_Surface, app: any) -> None:
        # TODO: lazy SDL_ConvertPixels
        self.destroyed = True
        if not surf:
            app.raise_error()
        self.app = app
        self.surface = surf
        self.format = PixelFormat(surf.contents.format.contents.format, app)
        self.w, self.h = surf.contents.w, surf.contents.h
        self.clip_rect = (
            surf.contents.clip_rect.x, surf.contents.clip_rect.y,
            surf.contents.clip_rect.w, surf.contents.clip_rect.h
        )
        self.destroyed = False
        self.color_mod = self.get_color_mod()
        self.color_key = self.get_color_key()
        self.alpha_mod = self.get_alpha_mod()
        self.blend_mode = self.get_blend_mode_int()
        try:
            self.has_color_key = bool(SDL_HasColorKey(self.surface))
        except NameError:
            self.has_color_key = False
        try:
            self.has_rle = bool(SDL_HasSurfaceRLE(self.surface))
        except NameError:
            self.has_rle = False
        self.must_lock = self.has_rle

    def update_blend_mode_by_alpha(self) -> None:
        SDL_SetSurfaceBlendMode(self.surface, SDL_BLENDMODE_NONE if self.alpha_mod >= 255 else SDL_BLENDMODE_NONE)

    def get_blend_mode(self) -> str:
        return self.app.r_blend_map[self.blend_mode]

    def set_blend_mode(self, blend_mode: str) -> None:
        self.set_blend_mode_int(self.app.blend_map[blend_mode])

    def get_blend_mode_int(self) -> int:
        try:
            blend_mode_ptr = ctypes.c_long()
            SDL_GetSurfaceBlendMode(self.surface, blend_mode_ptr)
        except ctypes.ArgumentError:
            blend_mode_ptr = ctypes.c_int()
            SDL_GetSurfaceBlendMode(self.surface, blend_mode_ptr)
        return blend_mode_ptr.value

    def set_blend_mode_int(self, blend_mode: int) -> None:
        SDL_SetSurfaceBlendMode(self.surface, blend_mode)
        self.blend_mode = self.get_blend_mode_int()

    def set_clip_rect(self, clip_rect: any) -> None:
        SDL_SetClipRect(self.surface, SDL_Rect(
            int(clip_rect[0]), int(clip_rect[1]), int(clip_rect[2]), int(clip_rect[3])
        ))
        self.clip_rect = (
            self.surface.contents.clip_rect.x, self.surface.contents.clip_rect.y,
            self.surface.contents.clip_rect.w, self.surface.contents.clip_rect.h
        )

    def lock(self) -> None:
        SDL_LockSurface(self.surface)

    def unlock(self) -> None:
        SDL_UnlockSurface(self.surface)

    def set_rle(self, enabled: bool) -> None:
        SDL_SetSurfaceRLE(self.surface, enabled)
        try:
            self.has_rle = bool(SDL_HasSurfaceRLE(self.surface))
        except NameError:
            self.has_rle = enabled
        self.must_lock = self.has_rle

    def map_rgba(self, color: any) -> int:
        return SDL_MapRGBA(self.surface.contents.format, int(color[0]), int(color[1]), int(color[2]), int(color[3]))

    def map_rgb(self, color: any) -> int:
        return SDL_MapRGB(self.surface.contents.format, int(color[0]), int(color[1]), int(color[2]))

    def set_alpha_mod(self, alpha_mod: int = 0) -> None:
        SDL_SetSurfaceAlphaMod(self.surface, alpha_mod)
        self.alpha_mod = self.get_alpha_mod()
        self.update_blend_mode_by_alpha()

    def get_alpha_mod(self) -> int:
        try:
            alpha_mod_ptr = ctypes.c_ulong()
            SDL_GetColorKey(self.surface, alpha_mod_ptr)
        except ctypes.ArgumentError:
            alpha_mod_ptr = ctypes.c_uint()
            SDL_GetColorKey(self.surface, alpha_mod_ptr)
        return alpha_mod_ptr.value

    def set_color_key(self, enabled: bool, color_key: int = 0) -> None:
        if enabled:
            self.color_key = color_key
        SDL_SetColorKey(self.surface, enabled, color_key)
        try:
            self.has_color_key = bool(SDL_HasColorKey(self.surface))
        except NameError:
            self.has_color_key = enabled

    def get_color_key(self) -> int:
        try:
            key_ptr = ctypes.c_ulong()
            SDL_GetColorKey(self.surface, key_ptr)
        except ctypes.ArgumentError:
            key_ptr = ctypes.c_uint()
            SDL_GetColorKey(self.surface, key_ptr)
        return key_ptr.value

    def set_color_mod(self, color_mod: any) -> None:
        SDL_SetSurfaceColorMod(self.surface, int(color_mod[0]), int(color_mod[1]), int(color_mod[2]))
        self.color_mod = self.get_color_mod()

    def get_color_mod(self) -> tuple:
        r_ptr, g_ptr, b_ptr = ctypes.c_ubyte(), ctypes.c_ubyte(), ctypes.c_ubyte()
        SDL_GetSurfaceColorMod(self.surface, r_ptr, g_ptr, b_ptr)
        return r_ptr.value, g_ptr.value, b_ptr.value

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        SDL_FreeSurface(self.surface)
        self.destroyed = True
        del self.app
        return False

    def copy(self) -> any:
        return Surface(SDL_DuplicateSurface(self.surface), self.app)

    def convert(self, pixel_format: PixelFormat) -> any:
        return Surface(SDL_ConvertSurfaceFormat(self.surface, pixel_format.pixel_format, 0), self.app)

    def fill_rect(self, color: any, fill_rect: any) -> None:
        SDL_FillRect(
            self.surface,
            fill_rect and SDL_Rect(int(fill_rect[0]), int(fill_rect[1]), int(fill_rect[2]), int(fill_rect[3])),
            self.map_rgba(color) if len(color) > 3 else self.map_rgb(color)
        )

    def fill_rects(self, color: any, fill_rects: any) -> None:
        SDL_FillRects(
            self.surface,
            (SDL_Rect * len(fill_rects))(*(fill_rect and SDL_Rect(
                int(fill_rect[0]), int(fill_rect[1]), int(fill_rect[2]), int(fill_rect[3])
            ) for fill_rect in fill_rects)),
            len(fill_rects),
            self.map_rgba(color) if len(color) > 3 else self.map_rgb(color)
        )

    def blit(self, src: any, src_rect: any = None, dst_rect: any = None) -> None:
        SDL_BlitSurface(
            src.surface,
            src_rect and SDL_Rect(int(src_rect[0]), int(src_rect[1]), int(src_rect[2]), int(src_rect[3])),
            self.surface,
            dst_rect and SDL_Rect(int(dst_rect[0]), int(dst_rect[1]), int(dst_rect[2]), int(dst_rect[3]))
        )

    def blit_to(self, dst: any, src_rect: any = None, dst_rect: any = None) -> None:
        SDL_BlitSurface(
            self.surface,
            src_rect and SDL_Rect(int(src_rect[0]), int(src_rect[1]), int(src_rect[2]), int(src_rect[3])),
            dst.surface,
            dst_rect and SDL_Rect(int(dst_rect[0]), int(dst_rect[1]), int(dst_rect[2]), int(dst_rect[3]))
        )

    def blit_scaled(self, src: any, src_rect: any = None, dst_rect: any = None) -> None:
        SDL_BlitScaled(
            src.surface,
            src_rect and SDL_Rect(int(src_rect[0]), int(src_rect[1]), int(src_rect[2]), int(src_rect[3])),
            self.surface,
            dst_rect and SDL_Rect(int(dst_rect[0]), int(dst_rect[1]), int(dst_rect[2]), int(dst_rect[3]))
        )

    def blit_scaled_to(self, dst: any, src_rect: any = None, dst_rect: any = None) -> None:
        SDL_BlitScaled(
            self.surface,
            src_rect and SDL_Rect(int(src_rect[0]), int(src_rect[1]), int(src_rect[2]), int(src_rect[3])),
            dst.surface,
            dst_rect and SDL_Rect(int(dst_rect[0]), int(dst_rect[1]), int(dst_rect[2]), int(dst_rect[3]))
        )

    def save_to_bmp(self, path: str) -> None:
        if SDL_SaveBMP(self.surface, self.app.stb(path)) < 0:
            self.app.raise_error()

    def save_to_png(self, path: str) -> None:
        if IMG_SavePNG(self.surface, self.app.stb(path)) < 0:
            self.app.raise_error(IMG_GetError)

    def save_to_jpg(self, path: str, quality: float = 1.0) -> None:
        if IMG_SaveJPG(self.surface, self.app.stb(path), int(quality * 100)) < 0:
            self.app.raise_error(IMG_GetError)

    def __del__(self) -> None:
        self.destroy()


class SurfaceAnimation:
    def __init__(self, animation: IMG_Animation, app: any) -> None:
        self.destroyed = True
        self.animation = animation
        self.w, self.h = animation.w, animation.h
        self.surfaces = tuple(Surface(animation.frames[_x], app) for _x in range(animation.count))
        self.delays = tuple(animation.delays[_x] for _x in range(animation.count))
        self.destroyed = False

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        IMG_FreeAnimation(self.animation)
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()



import math
import ctypes
from sdl2 import *


AAlevels = 256
AAbits = 8
DEFAULT_ELLIPSE_OVERSCAN = 4


def pixel(renderer: SDL_Renderer, x: int, y: int) -> int:
    return SDL_RenderDrawPoint(renderer, x, y)


def pixelRGBA(renderer: SDL_Renderer, x: int, y: int, r: int, g: int, b: int, a: int) -> int:
    return SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
        SDL_SetRenderDrawColor(renderer, r, g, b, a) | \
        SDL_RenderDrawPoint(renderer, x, y)


def pixelRGBAWeight(renderer: SDL_Renderer, x: int, y: int, r: int, g: int, b: int, a: int, weight: int) -> int:
    ax = ((a * weight) >> 8)
    if ax > 255:
        a = 255
    else:
        a = ax & 0x000000ff
    return pixelRGBA(renderer, x, y, r, g, b, a)


def hline(renderer: SDL_Renderer, x1: int, x2: int, y: int) -> int:
    return SDL_RenderDrawLine(renderer, x1, y, x2, y)


def hlineRGBA(renderer: SDL_Renderer, x1: int, x2: int, y: int, r: int, g: int, b: int, a: int) -> int:
    return SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
        SDL_SetRenderDrawColor(renderer, r, g, b, a) | \
        SDL_RenderDrawLine(renderer, x1, y, x2, y)


def vline(renderer: SDL_Renderer, x: int, y1: int, y2: int) -> int:
    return SDL_RenderDrawLine(renderer, x, y1, x, y2)


def vlineRGBA(renderer: SDL_Renderer, x: int, y1: int, y2: int, r: int, g: int, b: int, a: int) -> int:
    return SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
        SDL_SetRenderDrawColor(renderer, r, g, b, a) | \
        SDL_RenderDrawLine(renderer, x, y1, x, y2)


def rectangleRGBA(renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int, a: int) -> int:
    if x1 == x2:
        if y1 == y2:
            return pixelRGBA(renderer, x1, y1, r, g, b, a)
        else:
            return vlineRGBA(renderer, x1, y1, y2, r, g, b, a)
    elif y1 == y2:
        return hlineRGBA(renderer, x1, x2, y1, r, g, b, a)

    if x1 > x2:
        x1, x2 = x2, x1
    if y1 > y2:
        y1, y2 = y2, y1

    return SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
        SDL_SetRenderDrawColor(renderer, r, g, b, a) | \
        SDL_RenderDrawRect(renderer, SDL_Rect(x1, y1, x2 - x1, y2 - y1))


def roundedRectangleRGBA(
        renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, rad: int, r: int, g: int, b: int, a: int
) -> int:
    if not renderer or rad < 0:
        return -1

    if rad <= 1:
        return rectangleRGBA(renderer, x1, y1, x2, y2, r, g, b, a)
    elif x1 == x2:
        if y1 == y2:
            return pixelRGBA(renderer, x1, y1, r, g, b, a)
        else:
            return vlineRGBA(renderer, x1, y1, y2, r, g, b, a)
    elif y1 == y2:
        return hlineRGBA(renderer, x1, x2, y1, r, g, b, a)

    if x1 > x2:
        x1, x2 = x2, x1
    if y1 > y2:
        y1, y2 = y2, y1

    w, h = x2 - x1, y2 - y1

    if rad * 2 > w:
        rad = w // 2
    if rad * 2 > h:
        rad = h // 2

    xx1 = x1 + rad
    xx2 = x2 - rad
    yy1 = y1 + rad
    yy2 = y2 - rad

    result = arcRGBA(renderer, xx1, yy1, rad, 180, 270, r, g, b, a) | \
             arcRGBA(renderer, xx2, yy1, rad, 270, 360, r, g, b, a) | \
             arcRGBA(renderer, xx1, yy2, rad, 90, 180, r, g, b, a) | \
             arcRGBA(renderer, xx2, yy2, rad, 0, 90, r, g, b, a)

    if xx1 <= xx2:
        result |= hlineRGBA(renderer, xx1, xx2, y1, r, g, b, a) | hlineRGBA(renderer, xx1, xx2, y2, r, g, b, a)
    if yy1 <= yy2:
        result |= vlineRGBA(renderer, x1, yy1, yy2, r, g, b, a) | vlineRGBA(renderer, x2, yy1, yy2, r, g, b, a)

    return result


def roundedBoxRGBA(
        renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, rad: int, r: int, g: int, b: int, a: int
) -> int:
    if not renderer or rad < 0:
        return -1

    if rad <= 1:
        return boxRGBA(renderer, x1, y1, x2, y2, r, g, b, a)
    elif x1 == x2:
        if y1 == y2:
            return pixelRGBA(renderer, x1, y1, r, g, b, a)
        else:
            return vlineRGBA(renderer, x1, y1, y2, r, g, b, a)
    elif y1 == y2:
        return hlineRGBA(renderer, x1, x2, y1, r, g, b, a)

    if x1 > x2:
        x1, x2 = x2, x1
    if y1 > y2:
        y1, y2 = y2, y1

    w, h = x2 - x1 + 1, y2 - y1 + 1
    r2 = rad + rad

    if r2 > w:
        rad = w // 2
        r2 = rad + rad
    if r2 > h:
        rad = h // 2

    x = x1 + rad
    y = y1 + rad
    dx = x2 - x1 - rad - rad
    dy = y2 - y1 - rad - rad

    result = SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
             SDL_SetRenderDrawColor(renderer, r, g, b, a)

    cx, cy = 0, rad
    ocx = ocy = 0xffff
    df = 1 - rad
    d_e = 3
    d_se = -2 * rad + 5
    first = True
    while cx <= cy or first:
        first = False
        xpcx = x + cx
        xmcx = x - cx
        xpcy = x + cy
        xmcy = x - cy
        if not ocy == cy:
            if cy > 0:
                ypcy = y + cy
                ymcy = y - cy
                result |= hline(renderer, xmcx, xpcx + dx, ypcy + dy)
                result |= hline(renderer, xmcx, xpcx + dx, ymcy)
            else:
                result |= hline(renderer, xmcx, xpcx + dx, y)
            ocy = cy
        if not ocx == cx:
            if cx > 0:
                ypcx = y + cx
                ymcx = y - cx
                result |= hline(renderer, xmcy, xpcy + dx, ymcx)
                result |= hline(renderer, xmcy, xpcy + dx, ypcx + dy)
            else:
                result |= hline(renderer, xmcy, xpcy + dx, y)
            ocx = cx
        if df < 0:
            df += d_e
            d_e += 2
            d_se += 2
        else:
            df += d_se
            d_e += 2
            d_se += 4
            cy -= 1
        cx += 1

    if dx > 0 and dy > 0:
        result |= boxRGBA(renderer, x1, y1 + rad + 1, x2, y2 - rad, r, g, b, a)

    return result


def boxRGBA(renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int, a: int) -> int:
    if x1 == x2:
        if y1 == y2:
            return pixelRGBA(renderer, x1, y1, r, g, b, a)
        else:
            return vlineRGBA(renderer, x1, y1, y2, r, g, b, a)
    elif y1 == y2:
        return hlineRGBA(renderer, x1, x2, y1, r, g, b, a)

    if x1 > x2:
        x1, x2 = x2, x1
    if y1 > y2:
        y1, y2 = y2, y1

    return SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
        SDL_SetRenderDrawColor(renderer, r, g, b, a) | \
        SDL_RenderFillRect(renderer, SDL_Rect(x1, y1, x2 - x1 + 1, y2 - y1 + 1))


def line(renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int) -> int:
    return SDL_RenderDrawLine(renderer, x1, y1, x2, y2)


def lineRGBA(renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int, a: int) -> int:
    return SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
        SDL_SetRenderDrawColor(renderer, r, g, b, a) | \
        SDL_RenderDrawLine(renderer, x1, y1, x2, y2)


def _aalineRGBA(
        renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int, a: int, draw_endpoint: int
) -> int:
    xx0 = x1
    yy0 = y1
    xx1 = x2
    yy1 = y2
    if yy0 > yy1:
        tmp = yy0
        yy0 = yy1
        yy1 = tmp
        tmp = xx0
        xx0 = xx1
        xx1 = tmp
    dx = xx1 - xx0
    dy = yy1 - yy0
    if dx >= 0:
        xdir = 1
    else:
        xdir = -1
        dx = -dx

    if dx == 0:
        if draw_endpoint:
            return vlineRGBA(renderer, x1, y1, y2, r, g, b, a)
        else:
            if dy > 0:
                return vlineRGBA(renderer, x1, yy0, yy0 + dy, r, g, b, a)
            else:
                return pixelRGBA(renderer, x1, y1, r, g, b, a)
    elif dy == 0:
        if draw_endpoint:
            return hlineRGBA(renderer, x1, x2, y1, r, g, b, a)
        else:
            if dy > 0:
                return vlineRGBA(renderer, xx0, xx0 + xdir * dx, y1, r, g, b, a)
            else:
                return pixelRGBA(renderer, x1, y1, r, g, b, a)
    elif dx == dy and draw_endpoint:
        return lineRGBA(renderer, x1, y1, x2, y2, r, g, b, a)

    erracc = 0
    intshift = 32 - AAbits
    wgtcompmask = AAlevels - 1
    result = pixelRGBA(renderer, x1, y1, r, g, b, a)
    if dy > dx:
        erradj = ((dx << 16) // dy) << 16
        x0pxdir = xx0 + xdir
        while dy - 1:
            erracctmp = erracc
            erracc += erradj
            if erracc <= erracctmp:
                xx0 = x0pxdir
                x0pxdir += xdir
            yy0 += 1
            wgt = (erracc >> intshift) & 255
            result |= pixelRGBAWeight(renderer, xx0, yy0, r, g, b, a, 255 - wgt)
            result |= pixelRGBAWeight(renderer, x0pxdir, yy0, r, g, b, a, wgt)
    else:
        erradj = ((dy << 16) // dx) << 16
        y0p1 = yy0 + 1
        while dx - 1:
            erracctmp = erracc
            erracc += erradj
            if erracc <= erracctmp:
                yy0 = y0p1
                y0p1 += 1
        xx0 += xdir
        wgt = (erracc >> intshift) & 255
        result |= pixelRGBAWeight(renderer, xx0, yy0, r, g, b, a, 255 - wgt)
        result |= pixelRGBAWeight(renderer, xx0, y0p1, r, g, b, a, wgt)

    if draw_endpoint:
        result |= pixelRGBA(renderer, x2, y2, r, g, b, a)

    return result


def aalineRGBA(renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int, a: int) -> int:
    return _aalineRGBA(renderer, x1, y1, x2, y2, r, g, b, a, 1)


def circleRGBA(renderer: SDL_Renderer, x: int, y: int, rad: int, r: int, g: int, b: int, a: int) -> int:
    return ellipseRGBA(renderer, x, y, rad, rad, r, g, b, a)


def arcRGBA(
        renderer: SDL_Renderer, x: int, y: int, rad: int, start: int, end: int, r: int, g: int, b: int, a: int
) -> int:
    if rad < 0:
        return -1
    if rad == 0:
        return pixelRGBA(renderer, x, y, r, g, b, a)

    start %= 360
    end %= 360
    while start < 0:
        start += 360
    while end < 0:
        end += 360
    start %= 360
    end %= 360
    startoct = start // 45
    endoct = end // 45
    _oct = startoct - 1
    drawoct = 0
    stopval_start = stopval_end = 0
    temp = 0.0

    first = True
    while not _oct == endoct or first:
        first = False
        _oct = (_oct + 1) % 8
        if _oct == startoct:
            dstart = float(start)
            if _oct == 0 or _oct == 3:
                temp = math.sin(dstart * math.pi / 180)
            elif _oct == 1 or _oct == 6:
                temp = math.cos(dstart * math.pi / 180)
            elif _oct == 2 or _oct == 5:
                temp = -math.cos(dstart * math.pi / 180)
            elif _oct == 4 or _oct == 7:
                temp = -math.sin(dstart * math.pi / 180)
            temp *= rad
            stopval_start = int(temp)
            if _oct % 2:
                drawoct |= 1 << _oct
            else:
                drawoct &= 255 - (1 << _oct)
        if _oct == endoct:
            dend = float(end)
            if _oct == 0 or _oct == 3:
                temp = math.sin(dend * math.pi / 180)
            elif _oct == 1 or _oct == 6:
                temp = math.cos(dend * math.pi / 180)
            elif _oct == 2 or _oct == 5:
                temp = -math.cos(dend * math.pi / 180)
            elif _oct == 4 or _oct == 7:
                temp = -math.sin(dend * math.pi / 180)
            temp *= rad
            stopval_end = int(temp)
            if startoct == endoct:
                if start > end:
                    drawoct = 255
                else:
                    drawoct &= 255 - (1 << _oct)
            elif _oct % 2:
                drawoct &= 255 - (1 << _oct)
            else:
                drawoct |= 1 << _oct
        elif not _oct == startoct:
            drawoct |= (1 << _oct)

    result = SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
             SDL_SetRenderDrawColor(renderer, r, g, b, a)

    cx, cy = 0, rad
    df = 1 - rad
    d_e = 3
    d_se = -2 * rad + 5
    first = True
    while cx <= cy or first:
        first = False
        ypcy = y + cy
        ymcy = y - cy
        if cx > 0:
            xpcx = x + cx
            xmcx = x - cx
            if drawoct & 4:
                result |= pixel(renderer, xmcx, ypcy)
            if drawoct & 2:
                result |= pixel(renderer, xpcx, ypcy)
            if drawoct & 32:
                result |= pixel(renderer, xmcx, ymcy)
            if drawoct & 64:
                result |= pixel(renderer, xpcx, ymcy)
        else:
            if drawoct & 96:
                result |= pixel(renderer, x, ymcy)
            if drawoct & 6:
                result |= pixel(renderer, x, ypcy)
        xpcy = x + cy
        xmcy = x - cy

        if cx > 0 and not cx == cy:
            ypcx = y + cx
            ymcx = y - cx
            if drawoct & 8:
                result |= pixel(renderer, xmcy, ypcx)
            if drawoct & 1:
                result |= pixel(renderer, xpcy, ypcx)
            if drawoct & 16:
                result |= pixel(renderer, xmcy, ymcx)
            if drawoct & 128:
                result |= pixel(renderer, xpcy, ymcx)
        elif cx == 0:
            if drawoct & 24:
                result |= pixel(renderer, xmcy, y)
            if drawoct & 129:
                result |= pixel(renderer, xpcy, y)

        if stopval_start == cx:
            if drawoct & (1 << startoct):
                drawoct &= 255 - (1 << startoct)
            else:
                drawoct |= (1 << startoct)
        if stopval_end == cx:
            if drawoct & (1 << endoct):
                drawoct &= 255 - (1 << endoct)
            else:
                drawoct |= (1 << endoct)

        if df < 0:
            df += d_e
            d_e += 2
            d_se += 2
        else:
            df += d_se
            d_e += 2
            d_se += 4
            cy -= 1
        cx += 1

    return result


def aacircleRGBA(renderer: SDL_Renderer, x: int, y: int, rad: int, r: int, b: int, g: int, a: int) -> int:
    return aaellipseRGBA(renderer, x, y, rad, rad, r, g, b, a)


def _drawQuadrants(renderer: SDL_Renderer, x: int, y: int, dx: int, dy: int, _f: int) -> int:
    if dx == 0:
        if dy == 0:
            result = pixel(renderer, x, y)
        else:
            ypdy = y + dy
            ymdy = y - dy
            if _f:
                result = vline(renderer, x, ymdy, ypdy)
            else:
                result = pixel(renderer, x, ypdy) | pixel(renderer, x, ymdy)
    else:
        xpdx = x + dx
        xmdx = x - dx
        ypdy = y + dy
        ymdy = y - dy
        if _f:
            result = vline(renderer, xpdx, ymdy, ypdy) | vline(renderer, xmdx, ymdy, ypdy)
        else:
            result = pixel(renderer, xpdx, ypdy) | pixel(renderer, xmdx, ypdy) | \
                     pixel(renderer, xpdx, ymdy) | pixel(renderer, xmdx, ymdy)
    return result


def _ellipseRGBA(
        renderer: SDL_Renderer, x: int, y: int, rx: int, ry: int, r: int, g: int, b: int, a: int, _f: int
) -> int:
    if rx < 0 or ry < 0:
        return -1
    result = SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
             SDL_SetRenderDrawColor(renderer, r, g, b, a)
    if rx == 0:
        if ry == 0:
            return pixel(renderer, x, y)
        else:
            return vline(renderer, x, y - ry, y + ry)
    elif ry == 0:
        return hline(renderer, x - rx, x + rx, y)

    rxi, ryi = rx, ry
    if rxi >= 512 or ryi >= 512:
        ellipseOverscan = DEFAULT_ELLIPSE_OVERSCAN // 4
    elif rxi >= 256 or ryi >= 256:
        ellipseOverscan = DEFAULT_ELLIPSE_OVERSCAN // 2
    else:
        ellipseOverscan = DEFAULT_ELLIPSE_OVERSCAN
    oldX = scrX = 0
    oldY = scrY = ryi
    result |= _drawQuadrants(renderer, x, y, 0, ry, _f)

    rxi *= ellipseOverscan
    ryi *= ellipseOverscan
    rx2 = rxi * rxi
    rx22 = rx2 + rx2
    ry2 = ryi * ryi
    ry22 = ry2 + ry2
    curX = 0
    curY = ryi
    deltaX = 0
    deltaY = rx22 * curY
    _error = ry2 - rx2 * ryi + rx2 // 4
    while deltaX <= deltaY:
        curX += 1
        deltaX += ry22
        _error += deltaX + ry2
        if _error > 0:
            curY -= 1
            deltaY -= rx22
            _error -= deltaY
        scrX = curX // ellipseOverscan
        scrY = curY // ellipseOverscan
        if (not scrX == oldX and scrY == oldY) or (not scrX == oldX and not scrY == oldY):
            result |= _drawQuadrants(renderer, x, y, scrX, scrY, _f)
            oldX = scrX
            oldY = scrY

    if curY > 0:
        curXp1 = curX + 1
        curYm1 = curY - 1
        _error = ry2 * curX * curXp1 + ((ry2 + 3) // 4) + rx2 * curYm1 * curYm1 - rx2 * ry2
        while curY > 0:
            curY -= 1
            deltaY -= rx22
            _error += rx2
            _error -= deltaY
            if _error <= 0:
                curX += 1
                deltaX += ry22
                _error += deltaX
            scrX = curX // ellipseOverscan
            scrY = curY // ellipseOverscan
            if (not scrX == oldX and scrY == oldY) or (not scrX == oldX and not scrY == oldY):
                oldY -= 1
                while oldY >= scrY:
                    result |= _drawQuadrants(renderer, x, y, scrX, oldY, _f)
                    if _f:
                        oldY = scrY - 1
                    oldY -= 1
                oldX = scrX
                oldY = scrY
        if not _f:
            oldY -= 1
            while oldY >= 0:
                result |= _drawQuadrants(renderer, x, y, scrX, oldY, _f)
                oldY -= 1

    return result


def ellipseRGBA(renderer: SDL_Renderer, x: int, y: int, rx: int, ry: int, r: int, g: int, b: int, a: int) -> int:
    return _ellipseRGBA(renderer, x, y, rx, ry, r, g, b, a, 0)


def filledCircleRGBA(renderer: SDL_Renderer, x: int, y: int, rad: int, r: int, g: int, b: int, a: int) -> int:
    return _ellipseRGBA(renderer, x, y, rad, rad, r, g, b, a, 1)


def aaellipseRGBA(renderer: SDL_Renderer, x: int, y: int, rx: int, ry: int, r: int, g: int, b: int, a: int) -> int:
    if rx < 0 or ry < 0:
        return -1
    if rx == 0:
        if ry == 0:
            return pixel(renderer, x, y)
        else:
            return vline(renderer, x, y - ry, y + ry)
    elif ry == 0:
        return hline(renderer, x - rx, x + rx, y)
    a2 = rx * rx
    b2 = ry * ry
    ds = 2 * a2
    dt = 2 * b2
    xc2 = 2 * x
    yc2 = 2 * y
    sab = math.sqrt(a2 + b2)
    od = round(sab * 0.01) + 1
    dxt = round(a2 / sab) + od
    t = 0
    s = -2 * a2 * ry
    d = 0
    xp = x
    yp = y - ry
    result = SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE if a == 255 else SDL_BLENDMODE_BLEND) | \
             pixelRGBA(renderer, xp, yp, r, g, b, a) | pixelRGBA(renderer, xc2 - xp, yp, r, g, b, a) | \
             pixelRGBA(renderer, xp, yc2 - yp, r, g, b, a) | pixelRGBA(renderer, xc2 - xp, yc2 - yp, r, g, b, a)
    for _i in range(dxt):
        i = _i + 1
        xp -= 1
        d += t - b2
        if d >= 0:
            ys = yp - 1
        elif d - s - a2 > 0:
            if 2 * d - s - a2 >= 0:
                ys = yp + 1
            else:
                ys = yp
                yp += 1
                d -= s + a2
                s += ds
        else:
            yp += 1
            ys = yp + 1
            d -= s + a2
            s += ds
        t -= dt
        if s:
            cp = abs(d) / abs(s)
            if cp > 1.0:
                cp = 1.0
        else:
            cp = 1.0
        weight = int(cp * 255)
        iweight = 255 - weight
        xx = xc2 - xp
        result |= pixelRGBAWeight(renderer, xp, yp, r, g, b, a, iweight)
        result |= pixelRGBAWeight(renderer, xx, yp, r, g, b, a, iweight)
        result |= pixelRGBAWeight(renderer, xp, ys, r, g, b, a, weight)
        result |= pixelRGBAWeight(renderer, xx, ys, r, g, b, a, weight)
        yy = yc2 - yp
        result |= pixelRGBAWeight(renderer, xp, yy, r, g, b, a, iweight)
        result |= pixelRGBAWeight(renderer, xx, yy, r, g, b, a, iweight)
        yy = yc2 - ys
        result |= pixelRGBAWeight(renderer, xp, yy, r, g, b, a, weight)
        result |= pixelRGBAWeight(renderer, xx, yy, r, g, b, a, weight)
    dyt = round(b2 / sab) + od
    for _i in range(dxt):
        i = _i + 1
        yp += 1
        d += s + a2
        if d <= 0:
            xs = xp + 1
        elif d + t - b2 < 0:
            if 2 * d + t - b2 <= 0:
                xs = xp - 1
            else:
                xs = xp
                xp -= 1
                d += t - b2
                t -= dt
        else:
            xp -= 1
            xs = xp - 1
            d += t - b2
            t -= dt
        s += ds
        if t:
            cp = abs(d) / abs(t)
            if cp > 1.0:
                cp = 1.0
        else:
            cp = 1.0
        weight = int(cp * 255)
        iweight = 255 - weight
        xx = xc2 - xp
        yy = yc2 - yp
        result |= pixelRGBAWeight(renderer, xp, yp, r, g, b, a, iweight)
        result |= pixelRGBAWeight(renderer, xx, yp, r, g, b, a, iweight)
        result |= pixelRGBAWeight(renderer, xp, yy, r, g, b, a, iweight)
        result |= pixelRGBAWeight(renderer, xx, yy, r, g, b, a, iweight)
        xx = xc2 - xs
        result |= pixelRGBAWeight(renderer, xs, yp, r, g, b, a, weight)
        result |= pixelRGBAWeight(renderer, xx, yp, r, g, b, a, weight)
        result |= pixelRGBAWeight(renderer, xs, yy, r, g, b, a, weight)
        result |= pixelRGBAWeight(renderer, xx, yy, r, g, b, a, weight)

    return result


def filledEllipseRGBA(renderer: SDL_Renderer, x: int, y: int, rx: int, ry: int, r: int, g: int, b: int, a: int) -> int:
    return _ellipseRGBA(renderer, x, y, rx, ry, r, g, b, a, 0)


def _pieRGBA(
        renderer: SDL_Renderer, x: int, y: int, rad: int,
        start: int, end: int, r: int, g: int, b: int, a: int, filled: int
) -> int:
    # TODO
    return -1


def pieRGBA(
        renderer: SDL_Renderer, x: int, y: int, rad: int, start: int, end: int, r: int, g: int, b: int, a: int
) -> int:
    return _pieRGBA(renderer, x, y, rad, start, end, r, g, b, a, 0)


def filledPieRGBA(
        renderer: SDL_Renderer, x: int, y: int, rad: int, start: int, end: int, r: int, g: int, b: int, a: int
) -> int:
    return _pieRGBA(renderer, x, y, rad, start, end, r, g, b, a, 1)


def trigonRGBA(
        renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, x3: int, y3: int, r: int, g: int, b: int, a: int
) -> int:
    return polygonRGBA(renderer, (ctypes.c_int16 * 3)(x1, x2, x3), (ctypes.c_int16 * 3)(y1, y2, y3), 3, r, g, b, a)


def aatrigonRGBA(
        renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, x3: int, y3: int, r: int, g: int, b: int, a: int
) -> int:
    return aapolygonRGBA(renderer, (ctypes.c_int16 * 3)(x1, x2, x3), (ctypes.c_int16 * 3)(y1, y2, y3), 3, r, g, b, a)


def filledTrigonRGBA(
        renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, x3: int, y3: int, r: int, g: int, b: int, a: int
) -> int:
    return filledPolygonRGBA(renderer, (ctypes.c_int16 * 3)(x1, x2, x3), (ctypes.c_int16 * 3)(y1, y2, y3), 3, r, g, b, a)


def polygon(renderer: SDL_Renderer, vx: any, vy: any, n: int) -> int:
    # TODO
    return -1


def polygonRGBA(renderer: SDL_Renderer, vx: any, vy: any, n: int, r: int, g: int, b: int, a: int) -> int:
    # TODO
    return -1


def aapolygonRGBA(renderer: SDL_Renderer, vx: any, vy: any, n: int, r: int, g: int, b: int, a: int) -> int:
    # TODO
    return -1


def filledPolygonRGBAMT(
        renderer: SDL_Renderer, vx: any, vy: any, n: int,
        r: int, g: int, b: int, a: int, polyInts: any, polyAllocated: any
) -> int:
    # TODO
    return -1


def filledPolygonRGBA(renderer: SDL_Renderer, vx: any, vy: any, n: int, r: int, g: int, b: int, a: int) -> int:
    return filledPolygonRGBAMT(renderer, vx, vy, n, r, g, b, a, None, None)


def _evaluateBezier(data: any, ndata: int, t: float) -> int:
    # TODO
    return -1


def bezierRGBA(renderer: SDL_Renderer, vx: any, vy: any, n: int, s: int, r: int, g: int, b: int, a: int) -> int:
    # TODO
    return -1


def thickLineRGBA(
        renderer: SDL_Renderer, x1: int, y1: int, x2: int, y2: int, width: int, r: int, g: int, b: int, a: int
) -> int:
    if not renderer or width < 1:
        return -1
    if x1 == x2 and y1 == y2:
        wh = width // 2
        return boxRGBA(renderer, x1 - wh, y1 - wh, x2 + width, y2 + width, r, g, b, a)
    if width == 1:
        return lineRGBA(renderer, x1, y1, x2, y2, r, g, b, a)
    dx = x2 - x1
    dy = y2 - y1
    _l = math.sqrt(dx * dx + dy * dy)
    ang = math.atan2(dx, dy)
    adj = 0.1 + 0.9 * abs(math.cos(2.0 * ang))
    wl2 = (width - adj) / (2.0 * _l)
    nx = dx * wl2
    ny = dy * wl2
    return filledPolygonRGBA(
        renderer, [x1 + ny, x1 - ny, x2 - ny, x2 + ny], [y1 - nx, y1 + nx, y2 + nx, y2 - nx], 4, r, g, b, a
    )



import os
import sys
import shutil
from setuptools import setup, find_packages
from goodgame import __version__ as version


def read_file(fn: str) -> str:
    f = open(os.path.join(cwd, fn), 'r', encoding=encoding)
    result = f.read()
    f.close()
    return result


encoding = 'utf-8'
cwd = os.path.dirname(__file__) or os.getcwd()

readme_content = read_file('README.md')
license_content = read_file('LICENSE.md')
requirements_content = read_file('requirements.txt')

if 'sdist' in sys.argv or 'bdist_wheel' in sys.argv:
    dist_path = os.path.join(cwd, 'dist')
    if os.path.isdir(dist_path):
        shutil.rmtree(dist_path)

setup(
    name='goodgame',
    author='Pixelsuft',
    url='https://github.com/Pyxelsuft/goodgame',
    project_urls={
        'Readme': 'https://github.com/Pyxelsuft/goodgame/blob/main/README.md',
        'Documentation': 'https://github.com/Pyxelsuft/goodgame/tree/main/docs',
        'Issue tracker': 'https://github.com/Pyxelsuft/goodgame/issues'
    },
    version=version,
    packages=find_packages(),
    license='MIT',
    description='A simple and nice pysdl2 wrapper.',
    long_description=readme_content,
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=tuple(x for x in requirements_content.split('\n') if x.strip()),
    python_requires='>=3.6',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Topic :: Games/Entertainment',
        'Topic :: Multimedia',
        'Topic :: Software Development',
        'Operating System :: Android',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: iOS',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12'
    ],
    zip_safe=True,
    py_modules=['goodgame'],
    package_dir={'': '.'},
    keywords='goodgame'
)



import ctypes
from sdl2 import *


class AudioSpec:
    def __init__(
            self,
            spec: SDL_AudioSpec = SDL_AudioSpec(0, 0, 0, 0),
            is_capture: bool = False,
            device_name: str = None
    ) -> None:
        self.spec = spec
        self.is_capture = is_capture
        self.device_name = device_name
        self.callback = spec.callback
        self.channels = spec.channels
        self.format = spec.format
        self.freq = spec.freq
        self.padding = spec.padding
        self.samples = spec.samples
        self.silence = spec.silence
        self.size = spec.size
        self.userdata = spec.userdata

    def as_spec(self) -> SDL_AudioSpec:
        spec = SDL_AudioSpec(self.freq, self.format, self.channels, self.samples)
        spec.callback = self.callback
        spec.channels = self.channels
        spec.format = self.format
        spec.freq = self.freq
        spec.padding = self.padding
        spec.samples = self.samples
        spec.silence = self.silence
        spec.size = self.size
        spec.userdata = self.userdata
        return spec


class WAV:
    def __init__(self, spec: AudioSpec, buf_ptr: any, size: int) -> None:
        self.destroyed = True
        self.spec = spec
        self.buf_ptr = buf_ptr
        self.size = size
        self.destroyed = False

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        SDL_FreeWAV(self.buf_ptr)
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()


class AudioDevice:
    def __init__(self, app: any, device_id: int, spec: AudioSpec) -> None:
        self.destroyed = True
        self.app = app
        self.id = device_id
        self.spec = spec
        self.destroyed = False
        # TODO:
        #  finish (but why we need it when Mixer exists?)

    def queue(self, data: any, size: int) -> None:
        if SDL_QueueAudio(self.id, data, size) < 0:
            self.app.raise_error()

    def lock(self) -> None:
        SDL_LockAudioDevice(self.id)

    def unlock(self) -> None:
        SDL_UnlockAudioDevice(self.id)

    def pause(self) -> None:
        SDL_PauseAudioDevice(self.id, 1)

    def resume(self) -> None:
        SDL_PauseAudioDevice(self.id, 0)

    def get_status(self) -> str:
        status = SDL_GetAudioDeviceStatus(self.id)
        if status == SDL_AUDIO_PLAYING:
            return 'playing'
        elif status == SDL_AUDIO_PAUSED:
            return 'paused'
        else:
            return 'stopped'

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        SDL_CloseAudioDevice(self.id)
        del self.app
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()


class AudioDeviceManager:
    def __init__(self, app: any) -> None:
        self.app = app
        self.current_driver = app.bts(SDL_GetCurrentAudioDriver())
        self.playback_devices = []
        self.recording_devices = []
        try:
            self.default_playback_info = self.get_default_playback_info()
            self.default_recording_info = self.get_default_recording_info()
        except (RuntimeError, NameError):
            self.default_playback_info = AudioSpec(is_capture=False)
            self.default_recording_info = AudioSpec(is_capture=True)
        self.get_playback_devices()
        self.get_recording_devices()

    def load_wav(self, path: str) -> WAV:
        spec_ptr = SDL_AudioSpec(0, 0, 0, 0)
        buf_ptr = ctypes.POINTER(ctypes.c_uint8)()
        len_ptr = ctypes.c_uint32()
        if not SDL_LoadWAV(self.app.stb(path), spec_ptr, buf_ptr, len_ptr):
            self.app.raise_error()
        return WAV(AudioSpec(spec_ptr, False), buf_ptr, len_ptr.value)

    def open_device(self, spec: AudioSpec) -> AudioDevice:
        spec_ptr = SDL_AudioSpec(0, 0, 0, 0)
        device_id = SDL_OpenAudioDevice(
            spec.device_name and self.app.stb(spec.device_name),
            spec.is_capture,
            spec.as_spec(),
            spec_ptr,
            SDL_AUDIO_ALLOW_ANY_CHANGE
        )
        if device_id == 0:
            self.app.raise_error()
        return AudioDevice(self.app, device_id, AudioSpec(spec_ptr, spec.is_capture, spec.device_name))

    def get_default_playback_info(self) -> AudioSpec:
        name_ptr = ctypes.c_char_p()
        spec = SDL_AudioSpec(0, 0, 0, 0)
        SDL_GetDefaultAudioInfo(name_ptr, spec, 0)
        return AudioSpec(spec, False, self.app.bts(name_ptr.value or b''))

    def get_default_recording_info(self) -> AudioSpec:
        name_ptr = ctypes.c_char_p()
        spec = SDL_AudioSpec(0, 0, 0, 0)
        SDL_GetDefaultAudioInfo(name_ptr, spec, 1)
        return AudioSpec(spec, True, self.app.bts(name_ptr.value or b''))

    def get_playback_devices(self) -> None:
        for i in range(SDL_GetNumAudioDevices(0)):
            device_name = self.app.bts(SDL_GetAudioDeviceName(i, 0) or b'')
            spec_ptr = SDL_AudioSpec(0, 0, 0, 0)
            try:
                SDL_GetAudioDeviceSpec(i, 0, spec_ptr)
            except NameError:
                pass
            self.playback_devices.append(AudioSpec(spec_ptr, False, device_name))

    def get_recording_devices(self) -> None:
        for i in range(SDL_GetNumAudioDevices(1)):
            device_name = self.app.bts(SDL_GetAudioDeviceName(i, 1) or b'')
            spec_ptr = SDL_AudioSpec(0, 0, 0, 0)
            try:
                SDL_GetAudioDeviceSpec(i, 1, spec_ptr)
            except NameError:
                pass
            self.playback_devices.append(AudioSpec(spec_ptr, True, device_name))



import ctypes
from sdl2 import *


class Joystick:
    def __init__(self, app: any, index: int) -> None:
        self.destroyed = True
        self.index = index
        try:
            self.player_index = SDL_JoystickGetDevicePlayerIndex(index)
        except NameError:
            self.player_index = -1
        self.name = app.bts(SDL_JoystickNameForIndex(index))
        try:
            self.path = app.bts(SDL_JoystickPathForIndex(index))
        except NameError:
            self.path = ''
        self.guid = SDL_JoystickGetDeviceGUID(index)
        try:
            self.vendor = SDL_JoystickGetDeviceVendor(index)
            self.product = SDL_JoystickGetDeviceProduct(index)
            self.product_version = SDL_JoystickGetDeviceProductVersion(index)
        except NameError:
            self.path = 0
        self.type = {
            SDL_JOYSTICK_TYPE_UNKNOWN: 'unknown',
            SDL_JOYSTICK_TYPE_GAMECONTROLLER: 'game_controller',
            SDL_JOYSTICK_TYPE_WHEEL: 'wheel',
            SDL_JOYSTICK_TYPE_ARCADE_STICK: 'arcade_stick',
            SDL_JOYSTICK_TYPE_FLIGHT_STICK: 'flight_stick',
            SDL_JOYSTICK_TYPE_DANCE_PAD: 'dance_pad',
            SDL_JOYSTICK_TYPE_GUITAR: 'guitar',
            SDL_JOYSTICK_TYPE_DRUM_KIT: 'drum_kit',
            SDL_JOYSTICK_TYPE_ARCADE_PAD: 'arcade_pad',
            SDL_JOYSTICK_TYPE_THROTTLE: 'throttle'
        }.get(SDL_JoystickGetDeviceType(index))
        self.hat_map = {
            SDL_HAT_CENTERED: 'centered',
            SDL_HAT_DOWN: 'down',
            SDL_HAT_LEFT: 'left',
            SDL_HAT_LEFTDOWN: 'left_down',
            SDL_HAT_LEFTUP: 'left_up',
            SDL_HAT_RIGHT: 'right',
            SDL_HAT_RIGHTDOWN: 'right_down',
            SDL_HAT_RIGHTUP: 'right_up',
            SDL_HAT_UP: 'up'
        }
        self.power_map = {
            SDL_JOYSTICK_POWER_UNKNOWN: 'unknown',
            SDL_JOYSTICK_POWER_EMPTY: 'empty',
            SDL_JOYSTICK_POWER_LOW: 'low',
            SDL_JOYSTICK_POWER_MEDIUM: 'medium',
            SDL_JOYSTICK_POWER_FULL: 'full',
            SDL_JOYSTICK_POWER_WIRED: 'wired',
            SDL_JOYSTICK_POWER_MAX: 'max'
        }
        self.firmware_version = 0
        self.serial = b''
        self.instance_id = SDL_JoystickGetDeviceInstanceID(index)
        self.joystick = SDL_Joystick(index)
        self.opened = False
        self.num_axes = 0
        self.num_balls = 0
        self.num_hats = 0
        self.num_buttons = 0
        self.has_led = False
        self.has_rumble = False
        self.has_rumble_triggers = False
        self.destroyed = False
        # TODO:
        #  GUID functions
        #  Virtual Joysticks if needed (it's for you, I'm lazy!!!!!!!!!!!)

    def send_effect(self, data: bytes) -> None:
        SDL_JoystickSendEffect(self.joystick, data, len(data))

    def set_led(self, color: any) -> None:
        SDL_JoystickSetLED(self.joystick, int(color[0]), int(color[1]), int(color[2]))

    def rumble(self, low_freq_rumble: int, high_freq_rumble: int, duration: float) -> None:
        SDL_JoystickRumble(self.joystick, low_freq_rumble, high_freq_rumble, int(duration * 1000))

    def rumble_triggers(self, low_freq_rumble: int, high_freq_rumble: int, duration: float) -> None:
        SDL_JoystickRumbleTriggers(self.joystick, low_freq_rumble, high_freq_rumble, int(duration * 1000))

    def update_info(self) -> None:
        self.opened = True
        self.num_axes = SDL_JoystickNumAxes(self.joystick)
        self.num_balls = SDL_JoystickNumBalls(self.joystick)
        self.num_hats = SDL_JoystickNumHats(self.joystick)
        self.num_buttons = SDL_JoystickNumButtons(self.joystick)
        try:
            self.has_led = bool(SDL_JoystickHasLED(self.joystick))
            self.has_rumble = bool(SDL_JoystickHasRumble(self.joystick))
            self.has_rumble_triggers = bool(SDL_JoystickHasRumbleTriggers(self.joystick))
        except NameError:
            pass
        try:
            self.firmware_version = SDL_JoystickGetFirmwareVersion(self.joystick)
            self.serial = SDL_JoystickGetSerial(self.joystick) or b''
        except NameError:
            return

    def get_power_level(self) -> str:
        return self.power_map[SDL_JoystickCurrentPowerLevel(self.joystick)]

    def get_button(self, button: int) -> bool:
        return bool(SDL_JoystickGetButton(self.joystick, button))

    def get_ball(self, ball: int) -> tuple:
        x_ptr, y_ptr = ctypes.c_int(), ctypes.c_int()
        SDL_JoystickGetBall(self.joystick, ball, x_ptr, y_ptr)
        return x_ptr.value, y_ptr.value

    def get_hat(self, hat: int) -> str:
        return self.hat_map[SDL_JoystickGetHat(self.joystick, hat)]

    def get_axis_initial_state(self, axis: int) -> any:
        state_ptr = ctypes.c_int16()
        return SDL_JoystickGetAxisInitialState(self.joystick, axis, state_ptr) and state_ptr.value

    def get_axis(self, axis: int) -> int:
        return SDL_JoystickGetAxis(self.joystick, axis)

    def is_attached(self) -> bool:
        return bool(SDL_JoystickGetAttached(self.joystick))

    def open(self) -> None:
        self.joystick = SDL_JoystickOpen(self.index)
        self.update_info()

    def open_from_instance_id(self) -> None:
        self.joystick = SDL_JoystickFromInstanceID(self.instance_id)
        self.update_info()

    def open_from_player_index(self) -> None:
        self.joystick = SDL_JoystickFromPlayerIndex(self.player_index)
        self.update_info()

    def set_player_index(self, player_index: int = -1) -> None:
        self.player_index = player_index
        SDL_JoystickSetPlayerIndex(self.joystick, player_index)

    def destroy(self) -> bool:
        if self.destroyed:
            return True
        if self.opened:
            self.opened = False
            SDL_JoystickClose(self.joystick)
        self.destroyed = True
        return False

    def __del__(self) -> None:
        self.destroy()
